/**********************************************************************
 
	Copyright (C) 2007- Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/


#include	"utils.h"
#include	"u_file.h"
#include	"fa_cache.h"
#include	"memory_debug.h"


void
fac_init(FAC_INFO * info)
{
	info->seek_point = 0;
	if ( info->long_file )
		info->file_size = u_lseek64(info->fd,0,SEEK_END);
	else	info->file_size = u_lseek(info->fd,0,SEEK_END);
	if ( info->total_entries == 0 )
		info->total_entries = FAC_DEFAULT_ENTRIES;
	info->entries_tree = 0;
	info->entries.prev = info->entries.next = &info->entries;
	info->clean_point = &info->entries;
	info->dirty_nos = 0;
}

int
fac_cmp(FAC_ENTRY * e1,FAC_ENTRY * e2)
{
	if ( e1->offset < e2->offset )
		return -1;
	if ( e1->offset > e2->offset )
		return 1;
	return 0;
}

void
fac_delete(FAC_ENTRY_HEADER * e)
{
	e->next->prev = e->prev;
	e->prev->next = e->next;
}

void
fac_insert(FAC_ENTRY_HEADER * e,FAC_ENTRY_HEADER * n)
{
	n->prev = e;
	n->next = e->next;
	n->prev->next = n;
	n->next->prev = n;
}

FAC_ENTRY *
get_entry(int *status,FAC_INFO * info,U_INTEGER64 offset)
{
FAC_ENTRY * ret;
FAC_ENTRY cond;
AVT_NODE * n1;
U_INTEGER64 len;
FAC_ENTRY * e;
	if ( offset > info->file_size )
		return 0;
	cond.offset = offset & (-(INTEGER64)FAC_SIZE);
	n1 = avt_search(info->entries_tree,&cond,fac_cmp);
	if ( n1 == 0 ) {
		ret = d_alloc(sizeof(*ret));
		memset(ret,0,sizeof(*ret));
		ret->offset = cond.offset;
		ret->create_time = get_xltime();
		ret->data = d_alloc(FAC_SIZE);
		n1 = d_alloc(sizeof(*n1));
		n1->data = ret;
		len = info->file_size - cond.offset;
		if ( len > FAC_SIZE )
			ret->len = FAC_SIZE;
		else	ret->len = len;
		*status = 1;
		info->entries_nos ++;
	}
	else {
		ret = n1->data;
		*status = 0;
		fac_delete(&ret->h);
	}
	fac_insert(&info->entries,&ret->h);
	ret->access_time = get_xltime();
	
	for ( ; info->entries_nos > info->total_entries ;  ) {
		for ( ; ; ) {
			if ( info->clean_point == &info->entries ) {
				info->clean_point = info->clean_point->prev;
			}
			e = (FAC_ENTRY*)info->clean_point;
			if ( e->dirty ) {
				info->clean_point = e->h.prev;
				continue;
			}
			break;
		}
		info->clean_point = e->h.next;
		fac_delete(&e->h);
		info->entries_nos --;
		n1 = avt_delete(&info->entries_tree,e,fac_cmp);
		if ( n1 == 0 )
			er_panic("get_entry");
		if( n1->data != (void*)e )
			er_panic("get_entry");
		d_f_ree(n1);
		d_f_ree(e->data);
		d_f_ree(e);
	}
	return ret;
}

int
fac_trace_func(AVT_NODE * a,void * w)
{
FAC_INFO * info;
FAC_ENTRY * e;
	if ( e->dirty == 0 )
		return 0;
	e = a->data;
	info = (FAC_INFO*)w;
	if ( info->long_file )
		u_lseek64(info->fd,e->offset,SEEK_SET);
	else	u_lseek(info->fd,e->offset,SEEK_SET);
	u_write(info->fd,e->data,e->len);
	e->dirty = 0;
	info->dirty_nos --;
	return 0;
}

void
fac_purge(FAC_INFO * info)
{
	avt_trace_from_small(info->entries_tree,fac_trace_func,info);
	info->clean_point = &info->entries;
	if ( info->dirty_nos )
		er_panic("fac_purge");
}

void
fac_close(FAC_INFO * info)
{
FAC_ENTRY * e;
AVT_NODE * n;
	fac_purge(info);
	for ( ; info->entries_tree ; ) {
		e = (FAC_ENTRY*)info->entries.next;
		fac_delete(&e->h);
		n = avt_delete(&info->entries_tree,e,fac_cmp);
		if ( n == 0 )
			er_panic("fac_close");
		if ( n->data != (void*)e )
			er_panic("fac_close");
		d_f_ree(n);
		d_f_ree(e->data);
		d_f_ree(e);
	}
}

int
fac_read(FAC_INFO * info,char * data,int len)
{
int ret;
int s;
U_INTEGER64 offset,this_ofs;
FAC_ENTRY * e;
int ptr;
int status;
	ret = 0;
	offset = info->seek_point;
	for ( ; len > 0 ; ) {
		this_ofs = offset & (-(INTEGER64)FAC_SIZE);
		e = get_entry(&status,info,this_ofs);
		if ( e == 0 ) {
			errno = EIO;
			break;
		}
		if ( status ) {
			if ( info->long_file ) {
				u_lseek64(info->fd,e->offset,SEEK_SET);
				if ( e->len )
					u_read(info->fd,e->data,e->len);
			}
			else {
				u_lseek(info->fd,e->offset,SEEK_SET);
				if ( e->len )
					u_read(info->fd,e->data,e->len);
			}
		}
		ptr = offset - this_ofs;
		if ( ptr > e->len ) {
			break;
		}
		if ( ptr + len > e->len )
			s = e->len - ptr;
		else	s = len;
		memcpy(data,&e->data[ptr],s);
		data += s;
		len -= s;
		offset += s;
		ret += s;
	}
	info->seek_point = offset;
	return ret;
}

int
fac_wirte(FAC_INFO * info,char * data,int len)
{
int ret;
int s;
U_INTEGER64 offset,this_ofs;
FAC_ENTRY * e;
int ptr;
int status;
int _ret;
	ret = 0;
	offset = info->seek_point;
	for ( ; len > 0 ; ) {
		this_ofs = offset & (-(INTEGER64)FAC_SIZE);
		e = get_entry(&status,info,this_ofs);
		if ( e == 0 ) {
			errno = EIO;
			break;
		}
		_ret = -1;
		ptr = offset - this_ofs;
		if ( ptr > e->len ) {
			goto load;
		}
		_ret = 0;
		if ( ptr + len > e->len )
			s = e->len - ptr;
		else	s = len;
		if ( ptr || s != e->len ) {
		load:
			if ( info->long_file ) {
				u_lseek64(info->fd,e->offset,SEEK_SET);
				if ( e->len )
					u_read(info->fd,e->data,e->len);
			}
			else {
				u_lseek(info->fd,e->offset,SEEK_SET);
				if ( e->len )
					u_read(info->fd,e->data,e->len);
			}
			if ( _ret )
				break;
		}
		if ( s ) {
			if ( e->dirty == 0 ) {
				info->dirty_nos ++;
				if ( info->total_entries * FAC_DIRTY_RATE <= info->dirty_nos )
					fac_purge(info);
			}
			e->dirty = 1;
		}
		memcpy(&e->data[ptr],data,s);
		data += s;
		len -= s;
		offset += s;
		ret += s;
	}
	info->seek_point = offset;
	if ( info->file_size > offset )
		info->file_size = offset;
	return ret;
}


INTEGER64 
fac_lseek(FAC_INFO * info,U_INTEGER64 ofs,int pos)
{
	switch ( pos ) {
	case SEEK_SET:
		break;
	case SEEK_CUR:
		ofs = info->seek_point + ofs;
		break;
	case SEEK_END:
		ofs = info->file_size + ofs;
		break;
	default:
		return -1;
	}
	if ( ofs < 0 )
		return -1;
	if ( ofs > info->file_size )
		return -1;
	info->seek_point = ofs;
	return ofs;
}


